%#ok<*NOPTS>

addpathx('./gt-transport');

supplies = [ 3.0, 5.0, 0.0, 0.0, 2.5 ]';
demands = [ 0.5, 1.0, 2.0, 3.0, 4.0 ]';
n = length(supplies);
D = randi([0, 10], n, n);

% Printing out input
fprintf('Supplies:\n');
supplies'
fprintf('Demands:\n');
demands'
fprintf('Ground Distance:\n');
D

% Solving using gt-transport
fprintf('*** RESULTS for gt-transport ***\n\n');
t_gt = tic;
[ cost, flow ] = gt_transport(supplies, demands, D);
fprintf('Optimal Transportation Plan (cost = %f):\n', cost);
full(flow)
fprintf('--- DONE in %fsec ---\n\n', toc(t_gt));

% Compare with the MATLAB general-purpose LP-solver
fprintf('*** RESULTS for MATLAB''s LP-Solver ***\n\n');
t_linprog = tic;
A_ineq = [ ];
b_ineq = [ ];
A_eq = [ kron(ones(1, n), speye(n, n)); kron(speye(n, n), ones(1, n)) ];
b_eq = sparse([ supplies; demands ]);
lb = zeros(n * n, 1);
f = reshape(D, n * n, 1);
[flow, cost]  = linprog(f, A_ineq, b_ineq, A_eq, b_eq, lb, [], []);
flow = reshape(flow, n, n);
fprintf('Optimal Transportation Plan (cost = %f):\n', cost);
full(flow)
fprintf('--- DONE in %fsec ---\n\n', toc(t_linprog));


